/**
   This program evaluates (interprets) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

       Prog ::= Exp
              | '(' 'prog' (Fun | Exp)+ Exp ')'

        Fun ::= '(' 'fun' '(' VARIABLE VARIABLE+ ')' Exp+ ')'  // a function declaration

        Exp ::= Apply
              | If
              | While
              | Var
              | Set
              | Print
              | BExp
              | AExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Apply ::= '(' VARIABLE Exp+ ')'   // function name followed by actual parameters

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp+ ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

        Set ::= '(' 'set' VARIABLE Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [-][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class Evaluate
{
   private static final int DEBUG = 1;

   private static Environment globalEnv;  // reference to the global environment object
   private static Environment env;        // reference to the environment chain

   public static Value eval(Tree tree) throws EvalException
   {
      return evaluateProg( tree );
   }//eval()


   // Evaluate a prog expression.
   private static Value evaluateProg(Tree tree) throws EvalException
   {
      Value result = null;

      // Instantiate the global environment
      // (it will always be at the end of the
      // environment chain).
      globalEnv = new Environment();
      env = globalEnv;

      // Check which kind of Prog we have.
      if ( ! tree.getElement().equals("prog") )
      {
         // Evaluate the single expression.
         result = evaluateExp( tree );
      }
      else
      {
         // Evaluate each Fun or Exp in the Prog.
         // Any Fun will have the side effect of putting
         // a function name in the global environment.
         // Any Var expressions will have the side effect
         // of putting a variable in the global environment.
         // Any Set expressions will have the side effect
         // of changing a value in the global environment.
         // Any Print expressions will have the side effect
         // of printing an output.
         for (int i = 0; i < tree.degree()-1; i++)
         {
            if ( tree.getSubTree(i).getElement().equals("fun") )
            {
               handleFun( tree.getSubTree(i) );
            }
            else
            {
               evaluateExp( tree.getSubTree(i) );
            }
         }

         // Evaluate the last expression and use its
         // value as the value of the prog expression.
         result = evaluateExp( tree.getSubTree(tree.degree() - 1) );
      }

      return result;
   }//evaluateProg()


   /**
      Handle a function definition. Notice that the return type
      is void since Fun isn't an expression.

      This method mutates the global environment object.
      The Value object put into the environment by this method
      has the tag "lambda" and its value field is a reference
      to the function's Fun AST.
   */
   private static void handleFun(Tree tree) throws EvalException
   {
      // Get the function name.
      String name = tree.getSubTree(0).getElement();

      // Check if this function has already been defined.
      if ( globalEnv.definedLocal(name) )  // runtime check
      {
         throw new EvalException("function already exists: " + name);
      }

      // Get the Fun branch of the AST (the function value).
      Tree lambda = tree;

      // Create a function Value object and
      // add a <name, lambda> pair to the global environment.
      globalEnv.add(name, new Value(lambda) );

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes
   }//handleFun()


   // Evaluate an expression.
   private static Value evaluateExp(Tree tree) throws EvalException
   {
      Value result = null;

      String node = tree.getElement();

      if ( node.equals("if") )
         result = evaluateIf( tree );
      else if ( node.equals("while") )
         result = evaluateWhile( tree );
      else if ( node.equals("var") )
         result = evaluateVar( tree );
      else if ( node.equals("set") )
         result = evaluateSet( tree );
      else if ( node.equals("print") )
         result = evaluatePrint( tree );
      else if ( node.equals("&&") || node.equals("||")
             || node.equals("!") )
         result = evaluateBExp(tree);  // boolean expression
      else if ( node.equals("<")  || node.equals(">")
             || node.equals("<=") || node.equals(">=")
             || node.equals("==") || node.equals("!=") )
         result = evaluateRExp(tree);  // relational operator
      else if ( node.equals("+") || node.equals("-")
             || node.equals("*") || node.equals("/")
             || node.equals("%") || node.equals("^") )
         result = evaluateAExp(tree);  // arithmetic expression
      else if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
            result = new Value( node.equals("true") );
         else if ( node.matches("^[-]*[0-9][0-9]*") )
            result = new Value( Integer.parseInt( node ) );
         else if ( env.defined(node) )  // a variable
            result = env.lookUp( node );
         else  // runtime check
            throw new EvalException("undefined variable: " + node + "\n");
      }
      else // must be a function application
         result = evaluateApply( tree );

      return result;
   }//evaluateExp()


   /**
      This method "applies" a function value to actual parameters.
      This method evaluates the body of the function in an environment
      that binds the actual parameter values (from this function application)
      to the formal parameters (from the function's definition).
   */
   private static Value evaluateApply(Tree tree) throws EvalException
   {
      // The apply's root node is the function's name,
/*1*/ String funName = tree.getElement();

      // Check if this function has been defined in the global environment
      // (but the function name may be in the local environment).
      if ( ! env.defined(funName) )  // runtime check
      {
         throw new EvalException("undefined function: " + funName);
      }

      // Look up the name in the environment.
      Value funValue = env.lookUp( funName );

      // Check that what we are applying really is a function.
      if ( ! funValue.tag.equals(Value.LAMBDA_TAG))  // runtime check
      {
         throw new EvalException("bad function value: " + funValue);
      }

      // Get a reference to the function's AST.
      Tree lambda = funValue.valueL;

      // Check that the number of actual parameters
      // is equal to the number of formal parameters.
      // (Actually, all we really need to know is that
      // the number of actual parameters is at least
      // the number of formal parameters.)
      if ( tree.degree() != lambda.getSubTree(0).degree() )  // runtime check
      {
         throw new EvalException("wrong number of parameters: " + tree);
      }

      // Create a new environment object that is "nested"
      // in the global environment (lexical scope).
      // This environment is used to bind actual parameter
      // values to formal parameter names.
/*2*/ Environment localEnv = new Environment(globalEnv, "Function Activation");

      // Bind, in the new environment object, the actual parameter
      // values to the formal parameter names.
/*3*/ for (int i = 0; i < tree.degree(); i++) // iterate through the actual parameters
      {
         // Evaluate an actual parameter value
         // using the current environment chain
         // (NOT the new local environment object)
   /*4*/ Value actualParamValue = evaluateExp( tree.getSubTree(i) );

         // Retrieve, from within the function's definition,
         // a formal parameter name.
   /*5*/ String formalParamName = lambda.getSubTree(0).getSubTree(i).getElement();

         // Bind, in the new local environment object, the actual
         // parameter value to a formal parameter name.
   /*6*/ localEnv.add(formalParamName, actualParamValue);
      }

      if (DEBUG > 0) System.out.println( localEnv + "\n" ); // for debugging purposes

      // Evaluate the body of the function definition using the
      // new environment (which contains the binding of the actual
      // parameter values to the function's formal parameter names).
/*7*/ Environment originalEnv = env;
/*8*/ env = localEnv;

/*9*/ // Evaluate each Exp in the body of the function (for its side effect).
      for (int i = 1; i < lambda.degree() - 1; i++)
      {
         evaluateExp( lambda.getSubTree(i) );
      }

      // Evaluate the last Exp in the function for the return value.
/*10*/Value result = evaluateExp( lambda.getSubTree(lambda.degree()-1) );

      // Finally, restore the environment chain.
/*11*/env = originalEnv;

      return result;
   }//evaluateApply()


   // Evaluate an if-expression.
   private static Value evaluateIf(Tree tree) throws EvalException
   {
      if ( 3 != tree.degree() )  // runtime check
      {
         throw new EvalException("incorrect conditional expression: " + tree + "\n");
      }

      Value result = null;

      // Evaluate the boolean expression.
      Value conditionalExp = evaluateExp( tree.getSubTree(0) );
      // Do a runtime type check to make sure we have a boolean value.
      if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
      {
         throw new EvalException("illegal boolean expression: " + tree);
      }

      // Use the boolean value to choose a branch of the if-expression.
      if ( conditionalExp.valueB )
      {
         result = evaluateExp( tree.getSubTree(1) );
      }
      else
      {
         result = evaluateExp( tree.getSubTree(2) );
      }

      return result;
   }//evaluateIf()


   // Evaluate a while-loop expression.
   private static Value evaluateWhile(Tree tree) throws EvalException
   {
      if ( tree.degree() < 2 )  // runtime check
      {
         throw new EvalException("incorrect while expression: " + tree + "\n");
      }

      Value result = null;

      // Evaluate the boolean expression.
      Value conditionalExp = evaluateExp( tree.getSubTree(0) );
      // Do a runtime type check to make sure we have a boolean value.
      if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
      {
         throw new EvalException("illegal boolean expression: " + tree);
      }

      // Use the boolean value to control the looping.
      while ( conditionalExp.valueB )
      {
         // Evaluate each Exp in the body of the loop (for its side effect).
         for (int i = 1; i < tree.degree(); i++)
         {
            evaluateExp( tree.getSubTree(i) );
         }

         // Re-evaluate the boolean expression.
         conditionalExp = evaluateExp( tree.getSubTree(0) );
         // Do a runtime type check to make sure we have a boolean value.
         if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
         {
            throw new EvalException("illegal boolean expression: " + tree);
         }
      }

      // Always return false for a while-loop expression.
      result =  new Value( false );

      return result;
   }//evaluateWhile()


   // Evaluate a var expression.
   private static Value evaluateVar(Tree tree) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = null;

      // Get the variable name.
      String variable = tree.getSubTree(0).getElement();

      // Check if this variable has already been
      // declared in the local environment.
      if ( env.definedLocal(variable) )  // runtime check
      {
         throw new EvalException("variable already declared: " + variable + "\n");
      }

      // Get, and then evaluate, the expression.
      Tree expr = tree.getSubTree(1);
      result = evaluateExp( expr );

      // Create the new, local, variable.
      env.add(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//evaluateVar()


   // Evaluate a set expression.
   private static Value evaluateSet(Tree tree) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = null;

      // Get the variable name.
      String variable = tree.getSubTree(0).getElement();

      // Check if this variable has already been
      // declared in either the local or global environment.
      if ( ! env.defined(variable) )  // runtime check
      {
         throw new EvalException("undefined variable: " + variable);
      }

      // Get, and then evaluate, the expression.
      Tree expr = tree.getSubTree(1);
      result = evaluateExp(expr);

      // Update this variable in the environment.
      env.update(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//evaluateSet()


   // Evaluate a print expression.
   private static Value evaluatePrint(Tree tree) throws EvalException
   {
      if ( 1 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = evaluateExp( tree.getSubTree(0) );

      System.out.print( result + "\n" );

      return result;
   }//evaluatePrint()


   // Evaluate a boolean expression.
   private static Value evaluateBExp(Tree tree) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value value = evaluateExp( tree.getSubTree(0) );
      if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
      {
         throw new EvalException("not a boolean expression: "
                                  + tree.getSubTree(0) + "\n");
      }
      result = value.valueB;

      if ( node.equals("&&") )
      {
         if ( 2 > tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            if (result)
            {
               value = evaluateExp( tree.getSubTree(i) );
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalException("not a boolean expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result = result && value.valueB;
            }
            else  // short circuit the evaluation of '&&'
            {
               result = false;
               break;
            }
         }
      }
      else if ( node.equals("||") )
      {
         if ( 2 > tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            if (! result)
            {
               value = evaluateExp( tree.getSubTree(i) );
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalException("not a boolean expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result = result || value.valueB;
            }
            else  // short circuit the evaluation of '||'
            {
               result = true;
               break;
            }
         }
      }
      else if ( node.equals("!") )
      {
         if ( 1 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = ! result;
      }

      return new Value( result );
   }//evaluateBExp()


   // Evaluate a relational expression (which is a kind of boolean expression).
   private static Value evaluateRExp(Tree tree) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      boolean result = false;

      String op = tree.getElement();

      Value valueL = evaluateExp( tree.getSubTree(0) );
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(0) + "\n");
      }

      Value valueR = evaluateExp( tree.getSubTree(1) );
      if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(1) + "\n");
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( op.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( op.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( op.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( op.equals(">=") )
      {
         result = resultL >= resultR;
      }
      else if ( op.equals("==") )
      {
         result = resultL == resultR;
      }
      else if ( op.equals("!=") )
      {
         result = resultL != resultR;
      }

      return new Value( result );
   }//evaluateRExp()


   // Evaluate an arithmetic expression.
   private static Value evaluateAExp(Tree tree) throws EvalException
   {
      int result = 0;

      String node = tree.getElement();

      Value valueL = evaluateExp( tree.getSubTree(0) );
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(0) + "\n");
      }
      int resultL = valueL.valueI;
      int resultR = 0;

      Value valueR = null;
      if ( tree.degree() >= 2 )
      {
         valueR = evaluateExp( tree.getSubTree(1) );
         if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
         {
            throw new EvalException("not a integer expression: "
                                     + tree.getSubTree(1) + "\n");
         }
         resultR = valueR.valueI;
      }

      if ( node.equals("+") )
      {
         if ( tree.degree() == 1 )
            result = resultL;
         else
         {
            result = resultL + resultR;

            for (int i = 2; i < tree.degree(); i++)
            {
               Value temp = evaluateExp( tree.getSubTree(i) );
               if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
               {
                  throw new EvalException("not a integer expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result += temp.valueI;
            }
         }
      }
      else if ( node.equals("-") )
      {
         if ( 2 < tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         if ( tree.degree() == 1 )
            result = -resultL;
         else
            result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         if ( 1 == tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         result = resultL * resultR;

         for (int i = 2; i < tree.degree(); i++)
         {
            Value temp = evaluateExp( tree.getSubTree(i) );
            if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
            {
               throw new EvalException("not a integer expression: "
                                        + tree.getSubTree(i) + "\n");
            }
            result *= temp.valueI;
         }
      }
      else if ( node.equals("/") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = resultL % resultR;
      }
      else if ( node.equals("^") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = (int)Math.pow(resultL, resultR);
      }

      return new Value( result );
   }//evaluateAExp()
}//Evaluate
